"""
Creating monthly climatologies from the formatted netcdfs

Author: Marcus Falls marcus.falls@bsc.es
Date: 11/02/2019
"""

from netCDF4 import Dataset
import numpy as np
import os
import datetime
import time
import struct

from math import sqrt
from datetime import datetime as DATETIME
from geopy import distance
import numpy.ma as ma
import pandas as pd

### Choose source and destination directories
#pathin='/gpfs/scratch/bsc32/bsc32786/bgcargo/testfiles/bbp700/' # testing
#pathout='/gpfs/scratch/bsc32/bsc32786/bgcargo/climatologies/'

pathin='/esarchive/obs/argo/bgc/daily/bbp700/'    # all data
pathout='/esarchive/obs/argo/bgc/climatologies/'

### Choose ocean model
#model='nemo'
model='pisces'

### Choose which grid to use. So far only works for pisces and nemo outputs
map='/esarchive/scratch/mgalitap/aumont/pisces-v2-aumont17/PISCES_1m_19600101_19601231_ptrc_T.nc' # which grid is being used
#map='/esarchive/obs/argo/bgc/climatologies/mesh_mask_nemo.Ec3.2_O1L75.nc'

### Choose a file based on the metadata (for generating results based on different sensors)
sensordata='/esarchive/obs/argo/bgc/sensors_meta.csv'

bathymetry=1

def make_ts(source, destination, mask, model, sensors):

    # inputs:
    #
    # source: path of input files
    # destination: where the netcdf will go
    # mask: netCDF file containing grid and mask data
    # model: name of model, for now either 'nemo' or 'pisces'
    # sensordata: a csv file containing info on sensors
    # output: a netcdf file of the climatologies

    def getmonths(start, end):
        months = [start]
        monthinds = [[] for i in range(12)]
        curr = start
        count = 0
        while curr != end:

            # for i in range(200):
            m = int(curr[4:])
            y = int(curr[:-2])
            monthinds[m - 1].append(count)
            count = count + 1
            m = m + 1
            if m == 13:
                m = 1
                y = y + 1
            curr = "%04d" % y + "%02d" % m
            months.append(curr)
        return months, monthinds

    def mindistcoords(lat, lon, latitudes, longitudes):
        s=time.time()
        mindist = 20000
        minI = 0
        minJ = 0

        (x,y)=np.shape(latitudes)

        # for I in range(len(latitudes[0])):
        #     dist = distance.distance((lat, lon), (latitudes[int(x/2)][I], longitudes[int(x/2)][I]))
        #     if dist < mindist:
        #         mindist = dist
        #         minI = I
        #
        # for J in range(len(latitudes[:, minI])):
        #     dist = distance.distance((lat, lon), (latitudes[J][minI], longitudes[J][minI]))
        #     if dist < mindist:
        #         mindist = dist
        #         minJ = J


        for I in range(len(latitudes[0])):
            for J in range(len(latitudes[:, 0])):
                if abs(lat-latitudes[J][I])<2 and abs(lon-longitudes[J][I])<2:
                    dist = distance.distance((lat, lon), (latitudes[J][I], longitudes[J][I]))
                    if dist < mindist:
                        mindist = dist
                        minI = I
                        minJ=J

        #print('new',minJ,minI,mindist,lat,lon,latitudes[minJ][minI],longitudes[minJ][minI])
        e = time.time()
        #print(e-s)
        return minJ, minI

    def makeinds(obs, ref):
        mids = []
        for i in range(len(ref) - 1):
            mids.append((ref[i + 1] + ref[i]) / 2)
        res = []
        ind = 0
        check = mids[ind]
        for i in obs:
            # print(i,check,ind)
            if np.isnan(i):
                #res.append(np.nan)
                break
            else:

                while i > check and ind != len(mids) - 1:
                    ind = ind + 1  ### change this line
                    check = mids[ind]
                res.append(ind)
        #print('obs',len(obs),obs)
        #print('ref',len(ref),ref)
        #print('res',len(res),res)
        return res

        ## number of months and determining variable from the filename

    def getsensors(file, category):
        # input: csv file of metadata containing info in sensors, heading of category of interest
        # output: dictionary of floats a categories
        data = pd.read_csv(file)

        sensorcategories = {}
        for i in range(len(data['floats'])):
            if data[category][i] != '0':
                if data[category][i] not in sensorcategories.keys():
                    sensorcategories[data[category][i]] = []
                sensorcategories[data[category][i]].append(data['floats'][i])
        return sensorcategories

    files=os.listdir(source)
    files.sort()
    #print(files)
    first=files[0]
    varname=first[:-10]
    smonth=first[-9:-3]
    fmonth=files[-1][-9:-3]
    months,monthinds=getmonths(smonth,fmonth)


    ## extracting the dimensions, latitudes, latitudes and levels from the mask file so that the output has the same attributes
    maskdata=Dataset(mask,'r')
    X=maskdata.dimensions['x'].size
    Y=maskdata.dimensions['y'].size
    if 'z' in maskdata.dimensions:
        Z = maskdata.dimensions['z'].size
    elif 'deptht' in maskdata.dimensions:
        Z=maskdata.dimensions['deptht'].size ## pisces=deptht, nemo=z
    #Z=1 #### delete when levels are known

    gridlats = np.array(maskdata.variables['nav_lat'])
    gridlons = np.array(maskdata.variables['nav_lon'])
    if 'nav_lev' in maskdata.variables:
        gridlvls = np.array(maskdata.variables['nav_lev']) ## pisces=deptht, nemo=z
    elif 'deptht' in maskdata.variables:
        gridlvls = np.array(maskdata.variables['deptht'])

    ### if the model netcdf has bathymetry data, this is used to create a mask
    depths = np.zeros((len(months), Z, Y, X))
    if model=='nemo':  ### nemo outputs use a variable called bathymetry, the depth at each gridpoint
        if 'mbathy' in maskdata.variables and bathymetry==1:
            ismask=1
            gridmask= np.array(maskdata.variables['mbathy'])[0]

            for t in range(len(months)):
                for x in range(X):
                    for y in range(Y):
                        depth = gridmask[y][x]
                        for z in range(len(gridlvls)):
                            if gridlvls[z] < depth:
                                depths[t, z, y, x] = 1
                            else:
                                break
            # for t in range(1,len(months)):
            #     depths[t, z, y, x] = depths[0, z, y, x]
        else:
            gridmask = None
            ismask=0

    if model=='pisces': ### pisces uses NaN for all land points
        if 'NCHL' in maskdata.variables and bathymetry==1: ## this may need to be changed

            ismask = 1
            gridmask = np.array(maskdata.variables['NCHL'])[0]

            #for t in range(len(months)):
            for x in range(X):
                for y in range(Y):
                    for z in range(len(gridlvls)):

                        if gridmask[z,y,x]!=100000002004087734272:
                            depths[0, z, y, x] = 1
                        else:
                            break
            for t in range(1,len(months)):
                depths[t, :, :, :] = depths[0, :, :, :]

            print('mask of land values created')
        else:
            gridmask = None
            ismask = 0
            print('no mask of land values created')

    bbp700categories=getsensors(sensors, 'BBP700_CAT')

    # inserting the data into the netcdf
    os.chdir(source)

    ## dictionary of array for means, number of measurements,
    bbp700s={}
    # for all data
    bbp700s['totalsall']=np.zeros((len(months),Z,Y,X))
    bbp700s['totalsall'][:, :, :, :] = np.nan
    bbp700s['numbersall'] = np.zeros((len(months), Z, Y, X))
    bbp700s['maxall'] = np.zeros((len(months), Z, Y, X))
    bbp700s['minall'] = np.zeros((len(months), Z, Y, X))
    bbp700s['maxall'][:, :, :, :] = np.nan
    bbp700s['minall'][:, :, :, :] = np.nan
    # for each sensor
    for key in bbp700categories.keys():
        bbp700s['totals' + key]=np.zeros((len(months),Z,Y,X))
        bbp700s['totals' + key][:, :, :, :] = np.nan
        bbp700s['numbers' + key] = np.zeros((len(months), Z, Y, X))
        bbp700s['max' + key] = np.zeros((len(months), Z, Y, X))
        bbp700s['min' + key] = np.zeros((len(months), Z, Y, X))
        bbp700s['max' + key][:, :, :, :] = np.nan
        bbp700s['min' + key][:, :, :, :] = np.nan

    #totalsall=np.zeros((len(months),Z,Y,X))
    #totalsall[:,:,:,:]=np.nan




    s = time.time()
    for month in range(len(months)):
        f=varname+'_'+str(months[month])+'.nc'
        print(str(month) + ' extracting')
        # if month==40:
        #     break
        if f in files:
            dataold = Dataset(f, 'r')

            ## exctracting  the old variable data.
            varold_adj = np.array(dataold.variables[varname+'_adjusted'])
            varold = np.array(dataold.variables[varname])
            varold_adj_qc = np.array(dataold.variables[varname+'_adjusted_qc'])
            varold_qc = np.array(dataold.variables[varname+'_qc'])

            # taking each element of the raw data, replacing with adjusted if there is any,
            # and replacing with NaN if the qc flag is 4

            for d in range(len(varold)):
                for fl in range(len(varold[d])):
                    if not np.isnan(varold[d][fl][0]):
                        for p in range(len(varold[d][fl])):
                        #if varold[d][fl][p]!=np.nan:
                            if not np.isnan(varold_adj[d][fl][p]):
                                if varold_adj_qc[d][fl][p]!=4:
                                    varold[d][fl][p]=varold_adj[d][fl][p]
                                else:
                                    varold[d][fl][p] = np.nan
                            else:
                                if varold_qc[d][fl][p] == 4:
                                    varold[d][fl][p] = np.nan

            ## rest of process starts here
            floatsold=np.array(dataold.variables['floats'])
            latsold = np.array(dataold.variables['latitude'])
            lonsold = np.array(dataold.variables['longitude'])
            presold = np.array(dataold.variables['pressure'])

            for day in range(len(varold)):
                for fl in range(len(varold[day])):
                    if not np.isnan(varold[day][fl][0]):
                        #s = time.time()
                        (indy,indx)=mindistcoords(latsold[day,fl], lonsold[day,fl], gridlats, gridlons) ### this needs sped up

                        inds=makeinds(presold[day][fl],gridlvls)
                        for meas in range(len(inds)):
                        #for meas in range(len(varold[day][fl])):
                            if not np.isnan(varold[day][fl][meas]):
                                lvl=inds[meas]

                                # all data
                                if np.isnan(bbp700s['totalsall'][month,lvl,indy,indx]):
                                    bbp700s['totalsall'][month, lvl, indy, indx] = varold[day][fl][meas]
                                    bbp700s['maxall'][month, lvl, indy, indx] = varold[day][fl][meas]
                                    bbp700s['minall'][month, lvl, indy, indx] = varold[day][fl][meas]
                                else:
                                    bbp700s['totalsall'][month,lvl,indy,indx]=bbp700s['totalsall'][month,lvl,indy,indx]+varold[day][fl][meas]
                                    if bbp700s['maxall'][month, lvl, indy, indx] < varold[day][fl][meas]:
                                        bbp700s['maxall'][month, lvl, indy, indx] = varold[day][fl][meas]
                                    if bbp700s['minall'][month, lvl, indy, indx] > varold[day][fl][meas]:
                                        bbp700s['minall'][month, lvl, indy, indx] = varold[day][fl][meas]
                                bbp700s['numbersall'][month,lvl,indy,indx]=bbp700s['numbersall'][month,lvl,indy,indx]+1

                                # data sorted by sensor type
                                for key in bbp700categories.keys():
                                    if floatsold[fl] in bbp700categories[key]:
                                        if np.isnan(bbp700s['totals'+key][month, lvl, indy, indx]):
                                            bbp700s['totals'+key][month, lvl, indy, indx] = varold[day][fl][meas]
                                            bbp700s['max' + key][month, lvl, indy, indx] = varold[day][fl][meas]
                                            bbp700s['min' + key][month, lvl, indy, indx] = varold[day][fl][meas]
                                        else:
                                            bbp700s['totals'+key][month, lvl, indy, indx] = bbp700s['totals'+key][month, lvl, indy, indx] + varold[day][fl][meas]
                                            if bbp700s['max' + key][month, lvl, indy, indx] < varold[day][fl][meas]:
                                                bbp700s['max' + key][month, lvl, indy, indx] = varold[day][fl][meas]
                                            if bbp700s['min' + key][month, lvl, indy, indx] > varold[day][fl][meas]:
                                                bbp700s['min' + key][month, lvl, indy, indx] = varold[day][fl][meas]
                                        bbp700s['numbers'+key][month, lvl, indy, indx] = bbp700s['numbers'+key][month, lvl, indy, indx] + 1


            dataold.close()

        print(str(month) + ' complete')
            ## rest of process above


    print('all data extracted and binned')

    bbp700s['totalsall']=bbp700s['totalsall']/bbp700s['numbersall']
    for key in bbp700categories.keys():
        bbp700s['totals'+key] = bbp700s['totals'+key] / bbp700s['numbers'+key]

    ## create dataset of time series and define global attributes

    os.chdir(destination)

    datanew = Dataset(varname +'_'+model+ '_ts.nc', 'w', format='NETCDF4_CLASSIC')

    datanew.title = 'Monthly climatology for '+varname+' in '+model+' grid'
    datanew.history = 'Created ' + datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
    datanew.description = 'Gridded time series'
    datanew.references = "http://www.argodatamgt.org/Documentation"
    datanew.sources = 'Argo floats maintained by the following institutions: aoml, bodc, coriolis, csio, csiro, incois, jma, kma, kordi, meds'

    print('Dataset ' + varname + '_' + model + '_ts.nc created')

    ## create dimensions
    x = datanew.createDimension('x', X)  # latitude in degrees. Currently 1deg resolution (maybe change to make more flexible?)
    y = datanew.createDimension('y', Y)  # longitude in degrees. Currently 1deg resolution (maybe change to make more flexible?)
    z = datanew.createDimension('z', Z) # pressure/depth. need to agree on a suitable range
    t = datanew.createDimension('t', len(months))  # resolution in months. Months with no data will be included as nan arrays.

    ## create variables

    if ismask==1:
        dmask = datanew.createVariable('dmask', np.float32, ('t','z','y','x'), fill_value=np.nan)
        dmask.description="sea=0,land/under sea=1"
        dmask[:,:,:,:]=depths

    lvl_pres = datanew.createVariable('lvl_pres', np.float32, ('z','y','x'), fill_value=np.nan)
    lvl_pres.description="the pressure at each gridpoint"

    bbp700sncdf={}
    bbp700sncdf[varname+'_all_mean']=datanew.createVariable(varname+'_all_mean', np.float32, ('t','z','y','x'), fill_value=np.nan)
    bbp700sncdf[varname + '_all_max'] = datanew.createVariable(varname + '_all_max', np.float32, ('t', 'z', 'y', 'x'),fill_value=np.nan)
    bbp700sncdf[varname + '_all_min'] = datanew.createVariable(varname + '_all_min', np.float32, ('t', 'z', 'y', 'x'),fill_value=np.nan)
    for key in bbp700categories.keys():
        bbp700sncdf[varname + '_'+ key + '_mean'] = datanew.createVariable(varname + '_'+ key + '_mean', np.float32, ('t', 'z', 'y', 'x'), fill_value=np.nan)
        bbp700sncdf[varname + '_' + key + '_max'] = datanew.createVariable(varname + '_' + key + '_max', np.float32, ('t', 'z', 'y', 'x'), fill_value=np.nan)
        bbp700sncdf[varname + '_' + key + '_min'] = datanew.createVariable(varname + '_' + key + '_min', np.float32, ('t', 'z', 'y', 'x'), fill_value=np.nan)

    #var = datanew.createVariable(varname, np.float32, ('t','z','y','x'), fill_value=np.nan)

    bbp700s['totalsallnew']=bbp700s['totalsall'][:,:,:,:]
    bbp700s['maxallnew'] = bbp700s['maxall'][:, :, :, :]
    bbp700s['minallnew'] = bbp700s['minall'][:, :, :, :]

    for key in bbp700categories.keys():

        bbp700s['totals'+key+'new'] = bbp700s['totals'+key][:,:,:,:]
        bbp700s['max'+key+'new'] = bbp700s['max'+key][:, :, :, :]
        bbp700s['min'+key+'new'] = bbp700s['min'+key][:, :, :, :]

    fill = 0
    print('land values set as '+str(fill))
    if ismask==1:

        bbp700s['totalsallnew'] = ma.masked_where(depths == False, bbp700s['totalsallnew'])
        bbp700s['maxallnew'] = ma.masked_where(depths == False, bbp700s['maxallnew'])
        bbp700s['minallnew'] = ma.masked_where(depths == False, bbp700s['minallnew'])

        ma.set_fill_value(bbp700s['totalsallnew'], fill)
        ma.set_fill_value(bbp700s['maxallnew'], fill)
        ma.set_fill_value(bbp700s['minallnew'], fill)

        bbp700s['totalsallnew'] = bbp700s['totalsallnew'].filled()
        bbp700s['maxallnew'] = bbp700s['maxallnew'].filled()
        bbp700s['minallnew'] = bbp700s['minallnew'].filled()

        for key in bbp700categories.keys():

            bbp700s['totals'+key+'new'] = ma.masked_where(depths == False, bbp700s['totals'+key+'new'])
            bbp700s['max' + key + 'new'] = ma.masked_where(depths == False, bbp700s['max' + key + 'new'])
            bbp700s['min' + key + 'new'] = ma.masked_where(depths == False, bbp700s['min' + key + 'new'])

            ma.set_fill_value(bbp700s['totals'+key+'new'], fill)
            ma.set_fill_value(bbp700s['max' + key + 'new'], fill)
            ma.set_fill_value(bbp700s['min' + key + 'new'], fill)

            bbp700s['totals'+key+'new'] = bbp700s['totals'+key+'new'].filled()
            bbp700s['max'+key+'new'] = bbp700s['max'+key+'new'].filled()
            bbp700s['min'+key+'new'] = bbp700s['min'+key+'new'].filled()

    bbp700sncdf[varname + '_all_mean'][:,:,:,:]=bbp700s['totalsallnew']
    bbp700sncdf[varname + '_all_max'][:, :, :, :] = bbp700s['maxallnew']
    bbp700sncdf[varname + '_all_min'][:, :, :, :] = bbp700s['minallnew']

    for key in bbp700categories.keys():
        bbp700sncdf[varname + '_' + key + '_mean'][:, :, :, :] = bbp700s['totals'+key+'new']
        bbp700sncdf[varname + '_' + key + '_max'][:, :, :, :] = bbp700s['max' + key + 'new']
        bbp700sncdf[varname + '_' + key + '_min'][:, :, :, :] = bbp700s['min' + key + 'new']

    latitude=datanew.createVariable('nav_lat', np.float32, ('y','x'), fill_value=np.nan)
    latitude[:,:] = gridlats
    latitude.axis = 'x'

    longitude = datanew.createVariable('nav_lon', np.float32, ('y', 'x'), fill_value=np.nan)
    longitude[:,:] = gridlons
    longitude.axis='y'

    level = datanew.createVariable('nav_lev', np.float32, ('z',), fill_value=np.nan)
    level[:] = gridlvls
    level.axis = 'z'

    times = datanew.createVariable('times', np.int32, ('t',))
    times[:] = months
    times.axis = 't'

    ## create climatologies

    dataclim = Dataset(varname +'_'+model+  '_climatology.nc', 'w', format='NETCDF4_CLASSIC')

    dataclim.title = 'Monthly climatology for '+varname
    dataclim.history = 'Created ' + datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
    dataclim.description = 'Gridded climatologies'
    dataclim.references = "http://www.argodatamgt.org/Documentation"
    dataclim.sources = 'Argo floats maintained by the following institutions: aoml, bodc, coriolis, csio, csiro, incois, jma, kma, kordi, meds'

    print('Dataset ' + varname + '_'+model+'_climatology.nc created')

    ## create dimensions
    x = dataclim.createDimension('x', X)  # latitude in degrees. Currently 1deg resolution (maybe change to make more flexible?)
    y = dataclim.createDimension('y', Y)  # longitude in degrees. Currently 1deg resolution (maybe change to make more flexible?)
    z = dataclim.createDimension('z', Z) # pressure/depth. need to agree on a suitable range
    t = dataclim.createDimension('t', 12)  # resolution in months. Months with no data will be included as nan arrays.

    ## create variables
    if ismask==1:
        dmask = dataclim.createVariable('dmask', np.float32, ('t','z','y','x'), fill_value=np.nan)
        dmask.description="sea=1,land=0"
        depths=depths[:12,:,:,:]
        dmask[:,:,:,:]=depths

    lvl_pres = dataclim.createVariable('lvl_pres', np.float32, ('z','y','x'), fill_value=np.nan)
    lvl_pres.description = "the pressure at each gridpoint"

    bbp700sncdf[varname + '_all_mean'] = dataclim.createVariable(varname + '_all_mean', np.float32, ('t','z','y','x'), fill_value=np.nan)
    bbp700sncdf[varname + '_all_max'] = dataclim.createVariable(varname + '_all_max', np.float32, ('t', 'z', 'y', 'x'), fill_value=np.nan)
    bbp700sncdf[varname + '_all_min'] = dataclim.createVariable(varname + '_all_min', np.float32, ('t', 'z', 'y', 'x'), fill_value=np.nan)
    for key in bbp700categories.keys():
        bbp700sncdf[varname + '_'+key + '_mean'] = dataclim.createVariable(varname + '_' + key + '_mean', np.float32, ('t', 'z', 'y', 'x'),fill_value=np.nan)
        bbp700sncdf[varname + '_' + key + '_max'] = dataclim.createVariable(varname + '_' + key + '_max', np.float32,('t', 'z', 'y', 'x'), fill_value=np.nan)
        bbp700sncdf[varname + '_' + key + '_min'] = dataclim.createVariable(varname + '_' + key + '_min', np.float32,('t', 'z', 'y', 'x'), fill_value=np.nan)

    for mnth in range(12):
        bbp700s['mnthtotalsall']=bbp700s['totalsallnew'][monthinds[mnth],:,:,:]
        bbp700sncdf[varname + '_all_mean'][mnth,:,:,:]=np.nanmean(bbp700s['mnthtotalsall'],axis=0)
        bbp700s['mnthmaxall']=bbp700s['maxallnew'][monthinds[mnth],:,:,:]
        bbp700sncdf[varname + '_all_max'][mnth,:,:,:]=np.nanmax(bbp700s['mnthmaxall'],axis=0)
        bbp700s['mnthminall']=bbp700s['minallnew'][monthinds[mnth],:,:,:]
        bbp700sncdf[varname + '_all_min'][mnth,:,:,:]=np.nanmin(bbp700s['mnthminall'],axis=0)
        for key in bbp700categories.keys():
            bbp700s['mnthtotals'+key] = bbp700s['totals'+key+'new'][monthinds[mnth], :, :, :]
            bbp700sncdf[varname + '_'+key + '_mean'][mnth, :, :, :] = np.nanmean(bbp700s['mnthtotals'+key], axis=0)
            bbp700s['mnthmax'+key] = bbp700s['max'+key+'new'][monthinds[mnth], :, :, :]
            bbp700sncdf[varname + '_'+key + '_max'][mnth, :, :, :] = np.nanmax(bbp700s['mnthmax'+key], axis=0)
            bbp700s['mnthmin'+key] = bbp700s['min'+key+'new'][monthinds[mnth], :, :, :]
            bbp700sncdf[varname + '_'+key + '_min'][mnth, :, :, :] = np.nanmin(bbp700s['mnthmin'+key], axis=0)

    latitude=dataclim.createVariable('nav_lat', np.float32, ('y','x'), fill_value=np.nan)
    latitude[:,:] = gridlats
    latitude.axis = 'x'
    longitude = dataclim.createVariable('nav_lon', np.float32, ('y', 'x'), fill_value=np.nan)
    longitude[:,:] = gridlons
    longitude.axis = 'y'
    level = dataclim.createVariable('nav_lev', np.float32, ('z',), fill_value=np.nan)
    level[:] = gridlvls
    level.axis = 'z'
    times = dataclim.createVariable('times', np.int32, ('t',))
    times[:] = range(1,13)
    times.axis = 't'

    return datanew, dataclim

ts=make_ts(pathin,pathout,map,model,sensordata)